import time
import cupy as cp
import numpy as np

def timer(func):
    def wrapper(*args, **kwargs):
        time_start = time.time()
        func(*args, **kwargs)
        time_spent = time.time() - time_start
        # print("%s 的执行时间是 %f s" % (func.__name__, time.time() - time_start))
        return time_spent
    return wrapper

@timer
def cp_add(a, b):
    # print(cp.add(a, b))
    return cp.add(a, b)

@timer
def np_add(a, b):
    return np.add(a, b)

@timer
def cp_prod(a, b):
    # print(cp.add(a, b))
    return cp.multiply(a, b)

@timer
def np_prod(a, b):
    return np.multiply(a, b)

if __name__ == '__main__':
    # 不同大小的矩阵在 numpy 和 cupy 上进行验证
    len = 1
    # while len <= 10:
    while len <= 1e15:
        cp_time_list = []
        np_time_list = []
        print(u"目前矩阵大小为%sx%s=%s" % (len, len, len*len))
        for i in range(10):
            x = cp.random.randint(0, 10, size=(len, len))
            cp_time_list.append(cp_add(x, x))
            # cp_time_list.append(cp_prod(x, x))
            x = cp.asnumpy(x)
            np_time_list.append(np_add(x, x))
            # np_time_list.append(np_prod(x, x))
        print("cupy 加法的平均执行时间是 %f s" % (np.mean(cp_time_list)))
        print("numpy 加法的平均执行时间是 %f s" % (np.mean(np_time_list)))
        len *= 2


